#include "main.h"
struct tasks *tctl;
struct timer *tts;

struct task *taskInit(void)
{
	struct segdesc *sd = (struct segdesc *) GDT_ADDR;
	struct memory *mem = (struct memory *) MEMORY_ADDR;
	struct task *tsys, *tidle;
	int i;
	tctl = (struct tasks *) memAlloc(mem, sizeof (struct tasks));
	for (i = 0; i < TASKS; i++) {
		tctl->task[i].flag = 0;
		tctl->task[i].sel = (3 + i) * 8;
		tctl->task[i].tss.ldtr = (3 + TASKS + i) * 8;
		segdescSet(sd + 3 + i, (int) &tctl->task[i].tss, 103, TSS);
		segdescSet(sd + 3 + TASKS + i, (int) &tctl->task[i].ldt, 15, LDT);
	}
	for (i = 0; i < TLEVELS; i++) {
		tctl->level[i].running = 0;
		tctl->level[i].now = 0;
	}

	tsys = taskAlloc();
	tsys->flag = 2;
	tsys->pri = 2;
	tsys->lv = 1;
	taskAdd(tsys);
	taskSwitch0();
	trLoad(tsys->sel);

	tidle = taskAlloc();
	tidle->tss.esp = memAlloc(mem, 65536) + 65536;
	tidle->tss.eip = (int) &taskIdle;
	tidle->tss.es = 1 * 8;
	tidle->tss.cs = 2 * 8;
	tidle->tss.ss = 1 * 8;
	tidle->tss.ds = 1 * 8;
	tidle->tss.fs = 1 * 8;
	tidle->tss.gs = 1 * 8;
	taskRun(tidle, TLEVELS - 1, 1);

	tts = timerAlloc();
	timerSet(tts, tsys->pri, 0);
	tctl->tfpu = 0;
	return tsys;
}

void taskSwitch(void)
{
	struct tlevel *tl = &tctl->level[tctl->now];
	struct task *nt, *t = tl->task[tl->now];
	tl->now++;
	if (tl->now == tl->running) {
		tl->now = 0;
	}
	if (tctl->chlv != 0) {
		taskSwitch0();
		tl = &tctl->level[tctl->now];
	}
	nt = tl->task[tl->now];
	timerSet(tts, nt->pri, 0);
	if (nt != t) {
		farjmp(0, nt->sel);
	}
	return;
}

struct task *taskAlloc(void)
{
	struct task *t;
	int i;
	for (i = 0; i < TASKS; i++) {
		if (tctl->task[i].flag == 0) {
			t = &tctl->task[i];
			t->flag = 1;
			t->tss.eflags = 0x00000202;	/* IF = 1 */
			t->tss.eax = 0;
			t->tss.ecx = 0;
			t->tss.edx = 0;
			t->tss.ebx = 0;
			t->tss.ebp = 0;
			t->tss.esi = 0;
			t->tss.edi = 0;
			t->tss.es = 0;
			t->tss.ds = 0;
			t->tss.fs = 0;
			t->tss.gs = 0;
			t->tss.iomap = 0x40000000;
			t->tss.ss0 = 0;
			t->fpu[0] = 0x037f;		/* CW(control word) */
			t->fpu[1] = 0x0000;		/* SW(status word) */
			t->fpu[2] = 0xffff;		/* TW(tag word) */
			for (i = 3; i < 27; i++) {
				t->fpu[i] = 0;
			}
			t->lbyte = 0;
			return t;
		}
	}
	return 0;
}

void taskRun(struct task *task, int lv, int pri)
{
	if (lv < 0) {
		lv = task->lv;
	}
	if (pri > 0) {
		task->pri = pri;
	}
	if (task->flag == 2 && task->lv != lv) {
		taskRemove(task);
	}
	if (task->flag != 2) {
		task->lv = lv;
		taskAdd(task);
	}
	tctl->chlv = 1;
	return;
}

void taskSleep(struct task *task)
{
	struct task *nt;

	if (task->flag == 2) {
		nt = taskNow();
		taskRemove(task);
		if (task == nt) {
			timerCancel(tts);
			taskSwitch0();
			nt = taskNow();
			timerSet(tts, nt->pri, 0);
			farjmp(0, nt->sel);
		}
	}
	return;
}

struct task *taskNow(void)
{
	struct tlevel *tl = &tctl->level[tctl->now];
	return tl->task[tl->now];
}

void taskAdd(struct task *task)
{
	struct tlevel *tl = &tctl->level[task->lv];
	tl->task[tl->running] = task;
	tl->running++;
	task->flag = 2;
	return;
}

void taskRemove(struct task *task)
{
	struct tlevel *tl = &tctl->level[task->lv];
	int i;
	for (i = 0; i < tl->running; i++) {
		if (tl->task[i] == task) {
			break;
		}
	}
	tl->running--;
	if (i < tl->now) {
		tl->now--;
	}
	if (tl->now >= tl->running) {
		tl->now = 0;
	}
	task->flag = 1;
	for (; i < tl->running; i++) {
		tl->task[i] = tl->task[i + 1];
	}
	return;
}

void taskSwitch0(void)
{
	int i;
	for (i = 0; i < TLEVELS; i++) {
		if (tctl->level[i].running > 0) {
			break;
		}
	}
	tctl->now = i;
	tctl->chlv = 0;
	return;
}

void taskIdle(void)
{
	for (;;) {
		hlt();
	}
}
